REBOL [
	Title: "Uniserve: Async HTTP protocol"
	Author: "SOFTINNOV / Nenad Rakocevic"
	Version: 0.9.4
	Date: 02/11/2004
]

do-cache uniserve-path/libs/headers.r
do-cache uniserve-path/libs/cookies.r

system/schemes/default/proxy: make system/schemes/default/proxy [
	auth-method: 'BASIC
]

install-protocol [
	name: 'HTTP
	port-id: 80
	verbose: 0
	
	prot-version: "1.1"
	
	proto-header: compose [	; headers management
		User-Agent		"Mozilla/4.7 [en] (Win98; I)" ;"Uniserve/HTTP client"
		Accept			"*/*"
		Host			(none)
		Content-Type 	(none)
		Content-Length	(none)
		Connection		(none)
	]
	
	send-request: has [uri][
		uri: any [
			all [server/target join server/path server/target]
			server/path
		]
		either all [
			in server 'post
			series? server/post
		][
			insert-port/custom server uri reduce ['post server/post]
		][
			insert-port server uri
		]
	]
	
	clean-user-data: func [su][
		clear su/headers
		su/content: su/status-line: su/protocol: su/code: su/in-buffer: none
		su/state: 'init
		su/in-expected: su/in-trigger: 0
		su/out-method: 'GET
		out-headers: copy proto-header
	]
	
	on-connected: does [
		set-modes server [
			receive-buffer-size: 65536
			send-buffer-size: 16384
		]
		server/user-data: context [
			queue: make block! 1
			headers: make block! 10
			content:
			status-line:
			protocol:
			code: none
			in-expected: 
			in-trigger: 0
			in-buffer: none
			out-method: 'GET
			out-headers: copy proto-header
			out-url:
			out-content: 
			out-request: none
			auth-type: 
			auth-user:
			auth-pass:
			auth-state:
			auth-port:
			auth-header:
			auth-save-events: none
			state: 'init
		]
		server/proxy: system/schemes/default/proxy
		either server/proxy/type [
			switch server/proxy/auth-method [
				BASIC	[
					server/user-data/auth-header: join "Basic " enbase 
						join server/proxy/user [#":" server/proxy/pass] 
					send-request
				]
				DIGEST	[]
				NTLM	[NTLM-auth-process]
			]	
		][
			send-request
		]
	]

	non-crlf: complement charset "^M^/"
	digit: charset "0123456789"
	alpha: charset [#"a" - #"z" #"A" - #"Z" "-_%"]
	alphanum: union alpha digit

	handle-chunk: func [su data /local encoded? value size b s][
		on-stream-in server data
		either encoded?: select su/headers 'Transfer-Encoding [
			either find encoded? "chunked" [
				if su/in-buffer [insert tail su/in-buffer data]
				parse/all any [su/in-buffer data][
					some [
						b: copy size some alphanum thru lf s: (
							if zero? size: to integer! debase/base size 16 [
								on-response server su
								return true
							]
							if size > length? s [
								if none? su/in-buffer [su/in-buffer: copy b]
								return true
							]
						)
						copy value size skip 2 skip s: (
							either su/content [
								insert tail su/content value
							][
								su/content: copy value
							]
							s: remove/part su/in-buffer s
						) :s
					]
				]
			][
				log/warn rejoin ["encoding " encoded? " not supported"]
			]
			true
		][false]
	]
	
	handle-redirection: func [su all? /local url][
		if all [
			url: select su/headers 'Location
			any [not su/in-expected all?]
		][	
			if any [
				none? :on-redirect
				on-redirect server su url 
			][
; parse-url and if not the same server, open a new connection (close the old one)
				either "close" = select su/headers 'Connection [
					if not find url "http" [url: rejoin ["http://" server/host url]]				
					open-port url server/locals/evt-saved
				][
					clean-user-data su
					insert-port server url
				]
			]
		]	
	]
	
	handle-closing: func [su /local state][
		if any [
			none? state: select su/headers 'Connection
			"close" = state
		][
			close-port server
		]
	]
	
	on-close-server: does [
		on-closed server
	]
		
	on-raw-received: func [data /local su name value EOL s e valid? all? redir?][
		su: server/user-data
		either su/state = 'body [
			if not handle-chunk su data [
				either su/content [
					insert tail su/content data
				][
					su/content: copy data
				]
			]
		][
			valid?: parse/all data [
				s: "HTTP/" copy value 3 skip (su/protocol: attempt [to decimal! value])
				skip
				copy value 3 skip (su/code: attempt [to integer! value])
				skip 
				any non-crlf e: (su/status-line: copy/part s e)
				[crlf (EOL: crlf) | lf (EOL: lf) | cr (EOL: cr)]
				any [
					EOL copy value to end (
						if not handle-chunk su value [su/content: value]
						su/state: 'body
					)
					| copy name to #":" 2 skip copy value to EOL EOL (
						insert tail su/headers w: attempt [to word! name]
						insert tail su/headers value
						on-header-line server w value
						if w = 'Content-Length [
							su/in-expected: attempt [to integer! value]
						]
						if find [Set-Cookie Set-Cookie2] w [
							cookies/store server/host server/path value
						]
					)
				]
			]
		]
		redir?: all [300 <= su/code su/code < 400]
		if all?: all [
			su/in-expected
			any [
				all [redir? none? su/content]
				all [su/content su/in-expected = length? su/content]
			]
		][
			if not redir? [
				on-response server su
				handle-closing su		;--check for chunked case
			]
		]		
		handle-redirection su all?
	]
	
	new-insert-port: func [
		port
		url [url! string! block!]
		blk [block! none!]
		/local pu data name value proxy? post? req oh
	][	
		proxy?: to logic! all [port/proxy/type = 'generic port/proxy/host]
		post?: all [blk 'post = first blk 2 <= length? blk ]
		pu: port/user-data
		req: make string! 256
		oh: pu/out-headers
		pu/out-url: any [all [proxy? port/url] url]
		if post? [
			pu/out-method: 'POST
			h-store oh 'Content-Type "application/x-www-form-urlencoded"
		]
		if data: cookies/build form port/host port/path [
			h-store oh 'Cookie data
		]
		if not proxy? [h-store oh 'Host form any [port/host port/remote-ip]]
		
		data: any [all [post? second blk] pu/out-content none]
		h-store oh 'Content-Length any [all [data form length? data] none]
		
		if pu/auth-header [
			h-store oh
				pick [Proxy-Authorization Authorization] proxy?
				pu/auth/header
		]
		
		on-sending port pu
		
		insert tail req pu/out-method
		insert tail req #" "
		insert tail req pu/out-url
		insert tail req " HTTP/"
		insert tail req prot-version
		insert tail req crlf
		foreach-nv-pair oh [
			if value [
				insert tail req name
				insert tail req ": "
				insert tail req value
				insert tail req crlf
			]
		]
		insert tail req crlf
		if verbose > 0 [log/info join "request=>^/" mold req]
		if data [
			insert tail req data
			if verbose > 1 [log/info join "posted data=>^/" mold data]
		]
		
		write-server/with req port
		pu/state: 'header
		pu/out-request: req
		on-sent port
	]

	events: [
		on-header-line		;-- [port [port!] name [word!] value [string!]]
		on-header			;-- [port [port!] response [object!]]
		on-content			;-- [port [port!] data [string!]]
		on-response			;-- [port [port!] response [object!]]
		on-redirect			;-- [port [port!] response [object!] url [string!]]
		on-sending			;-- [port [port!] request [object!]]
		on-sent				;-- [port [port!]]
		on-stream-in		;-- [port [port!] data [string!]]
		on-stream-out		;-- [port [port!] data [string!]]
		on-closed			;-- [port [port!]]
		on-error			;-- [port [port!] code [word!]]		; port should be the first arg !
	]
]

{

p: open-port/with http://... [
	on-sending: func [port out][
		out/method: 'HEAD
	]
	on-response: func [port resp url queue][
	
	]
][
	targets: [http://... http://... ...]
]

insert-port p [http://... http://... ...]


send http://... 'GET [
	events
]
send http://... 'HEAD [
	events
]
send http://... [POST data][
	events
]

send http://... 


}